# -*- coding: utf-8 -*-
import numpy as np
from ..comms import printer
from ..tests import ductile
from ..tests import eigenvals
from ..utils.constants import *
from ..utils.elements import ELEMENTS
from ..utils.crystalutils import *
from .structure import Structure


class ElasticProperties:
    def __init__(self, elastic_tensor, structure=None, crystal_type=None):
        self.elastic_tensor = np.matrix(elastic_tensor)
        self.compaliance_tensor = self.elastic_tensor.I
        self.structure = structure
        self.crystal_type = crystal_type
        
        if crystal_type is not None or structure is not None:
            crystal_select(
                cnew=self.elastic_tensor,
                cell=self.structure.spglib_cell,
                crystal_type=self.crystal_type,
            )

    @property
    def K_v(self):
        """

        Returns
        -------
        KV : float
            Bulk Modulus Voigt.

        """
        cnew = self.elastic_tensor
        K_v = (cnew[0, 0] + cnew[1, 1] + cnew[2, 2]) + 2 * (
            cnew[0, 1] + cnew[1, 2] + cnew[2, 0]
        )
        K_v = K_v / 9.0
        return K_v

    @property
    def bulk_modulus_voigt(self):
        """


        Returns
        -------
        float
            Bulk Modulus Voigt.

        """
        return self.K_v

    @property
    def G_v(self):
        """


        Returns
        -------
        G_v : float
            Shear Modulus Voigt.

        """
        cnew = self.elastic_tensor
        ## Shear: Voigt
        G_v = (
            (cnew[0, 0] + cnew[1, 1] + cnew[2, 2])
            - (cnew[0, 1] + cnew[1, 2] + cnew[2, 0])
            + 3 * (cnew[3, 3] + cnew[4, 4] + cnew[5, 5])
        )
        G_v = G_v / 15.0
        return G_v

    @property
    def shear_modulus_voight(self):
        """


        Returns
        -------
        float
            Shear Modulus Voigt.

        """
        return self.G_v

    @property
    def E_v(self):
        """


        Returns
        -------
        float
            Young's Modulus.

        """
        return (9 * self.K_v * self.G_v) / (3 * self.K_v + self.G_v)

    @property
    def youngs_modulus_voigt(self):
        """


        Returns
        -------
        float
            Young's Modulus Voigt.

        """
        return self.E_v

    @property
    def Nu_v(self):
        """


        Returns
        -------
        float
            Poisson's Ratio Voigt.

        """
        return (3 * self.K_v - self.E_v) / (6 * self.K_v)

    @property
    def poissons_ratio_voigt(self):
        """


        Returns
        -------
        float
            Poisson's Ratio Voigt.

        """
        return self.Nu_v

    @property
    def M_v(self):
        """


        Returns
        -------
        float
            P-wave modulus voigt.

        """
        return self.K_v + (4 * self.G_v / 3.0)

    @property
    def p_wave_modulus_voigt(self):
        """


        Returns
        -------
        float
            P-wave modulus voigt.

        """
        return self.M_v

    @property
    def K_r(self):
        """


        Returns
        -------
        K_r : float
            Bulk Modulus Reuss.

        """
        snew = self.compaliance_tensor
        K_r = (snew[0, 0] + snew[1, 1] + snew[2, 2]) + 2 * (
            snew[0, 1] + snew[1, 2] + snew[2, 0]
        )
        K_r = 1.0 / K_r
        return K_r

    @property
    def bulk_modulus_reuss(self):
        """


        Returns
        -------
        float
            Bulk Modulus Reuss.

        """
        return self.K_r

    @property
    def G_r(self):
        """


        Returns
        -------
        G_r : float
            Shear Modulus Reuss.

        """
        snew = self.compaliance_tensor
        G_r = (
            4 * (snew[0, 0] + snew[1, 1] + snew[2, 2])
            - 4 * (snew[0, 1] + snew[1, 2] + snew[2, 0])
            + 3 * (snew[3, 3] + snew[4, 4] + snew[5, 5])
        )
        G_r = 15.0 / G_r
        return G_r

    @property
    def shear_modulus_reuss(self):
        """


        Returns
        -------
        G_r : float
            Shear Modulus Reuss.

        """
        return self.G_r

    @property
    def E_r(self):
        """


        Returns
        -------
        float
            Young's Modulus Reuss.

        """
        return (9 * self.K_r * self.G_r) / (3 * self.K_r + self.G_r)

    @property
    def youngs_modulus_reuss(self):
        """


        Returns
        -------
        float
            Young's modulus Reuss.

        """
        return self.E_r

    @property
    def Nu_r(self):
        """


        Returns
        -------
        float
            Poisson's Ration Reuss.

        """
        return (3 * self.K_r - self.E_r) / (6 * self.K_r)

    @property
    def poissons_ratio_reuss(self):
        """


        Returns
        -------
        float
            Poisson's Ration Reuss.

        """
        return self.Nu_r

    @property
    def M_r(self):
        """


        Returns
        -------
        float
            P-Wave Modulus Reuss.

        """
        return self.K_r + (4 * self.G_r / 3.0)

    @property
    def p_wave_modulus_reuss(self):
        """


        Returns
        -------
        float
            P-Wave Modulus Reuss.

        """
        return self.M_r

    @property
    def K_vrh(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation Bulk Modulus.

        """
        return (self.K_r + self.K_v) / 2

    @property
    def bulk_modulus_voigt_reuss_hill(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation Bulk Modulus.

        """
        return self.K_vrh

    @property
    def G_vrh(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation Shear Modulus.

        """
        return (self.G_v + self.G_r) / 2

    @property
    def shear_modulus_voight_reuss_hill(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation Shear Modulus.

        """
        return self.G_vrh

    @property
    def E_vrh(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation Young's Modulus.

        """
        return (self.E_v + self.E_r) / 2

    @property
    def youngs_modulus_voigt_reuss_hill(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation Young's Modulus.

        """
        return self.E_vrh

    @property
    def Nu_vrh(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation Poisson's Modulus.

        """
        return (self.Nu_v + self.Nu_r) / 2.0

    @property
    def poissons_ratio_voigt_reuss_hill(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation Poisson's Modulus.

        """
        return self.Nu_vrh

    @property
    def M_vrh(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation P-wave Modulus.

        """
        return (self.M_v + self.M_r) / 2.0

    @property
    def p_wave_modulus_voigt_reuss_hill(self):
        """


        Returns
        -------
        float
            Voigt-Reuss-Hill Approximation P-wave Modulus.

        """
        return self.M_vrh

    @property
    def KG_ratio_v(self):
        """


        Returns
        -------
        float
            Bulk/Shear ration voigt.

        """
        return self.K_v / self.G_v

    @property
    def KG_ratio_r(self):
        """


        Returns
        -------
        float
            Bulk/Shear ration reuss.

        """
        return self.K_r / self.G_r

    @property
    def KG_ratio_vrh(self):
        """


        Returns
        -------
        float
            Bulk/Shear ration voigt reuss hill.

        """
        return self.K_vrh / self.G_vrh

    @property
    def A_z(self):
        """


        Returns
        -------
        float
            Zenner Anisotropy only for Cubic Crystals.

        """

        cnew = self.elastic_tensor
        return 2 * cnew[3, 3] / (cnew[0, 0] - cnew[0, 1])

    @property
    def anisotropy_zenner(self):
        """


        Returns
        -------
        float
            Zenner Anisotropy only for Cubic Crystals.

        """
        return self.anisotropy_zenner

    @property
    def A_u(self):
        """


        Returns
        -------
        float
            Ranganathan and Ostoja-Starzewski method: Phys. Rev. Lett. 101, 055504 (2008).
            for any crystalline symmetry: Universal anisotropy index.
            Note: AU is a relative measure of anisotropy with respect to a limiting value.
            For example, AU does not prove that a crystal having AU = 3 has double the anisotropy
            of another crystal with AU = 1.5. I""'

        """
        return (self.K_v / self.K_r) + 5 * (self.G_v / self.G_r) - 6.0

    @property
    def anisotropy_universal(self):
        """


        Returns
        -------
        float
            Ranganathan and Ostoja-Starzewski method: Phys. Rev. Lett. 101, 055504 (2008).
            for any crystalline symmetry: Universal anisotropy index
            Note: AU is a relative measure of anisotropy with respect to a limiting value.
            For example, AU does not prove that a crystal having AU = 3 has double the anisotropy
            of another crystal with AU = 1.5. I""'

        """
        return self.A_u

    @property
    def A_l(self):
        """


        Returns
        -------
        float
               log-Euclidean anisotropy parameter by Christopher M. Kube, AIP Advances 6, 095209 (2016)
               AL  CV , CR   is based on the distance between the averaged stiffnesses CV and
               CR , which is more appropriate. Clearly, AL  CV , CR   is zero when the crystallite is isotropic.

        """
        return np.sqrt(5) * 2.303 * np.log(1 + (self.A_u / 5))

    @property
    def ductility(self):
        """


        Returns
        -------
        string
            The ductility of the material

        """
        return ductile.ductile_test(self.KG_ratio_vrh)

    @property
    def anisotropy_log_euclidean(self):
        """


        Returns
        -------
        float
               log-Euclidean anisotropy parameter by Christopher M. Kube, AIP Advances 6, 095209 (2016)
               AL  CV , CR   is based on the distance between the averaged stiffnesses CV and
               CR , which is more appropriate. Clearly, AL  CV , CR   is zero when the crystallite is isotropic.

        """
        return self.A_l

    @property
    def velocity_transverse(self):
        """


        Returns
        -------
        float
            Transverse Sound Velocity(m/s) from Navier's equation using Voigt-Reuss-Hill Approximation

        """
        if self.structure is None:
            raise Exception("No structure was provided")
            
        G = self.G_vrh * 1.0e9
        return np.sqrt((G / self.structure.density))

    @property
    def velocity_logitudinal(self):
        """


        Returns
        -------
        float
            longitudinal Sound velocity(m/s) from Navier's equation using Voigt-Reuss-Hill Approximation

        """
        if self.structure is None:
            raise Exception("No structure was provided")
        G = self.G_vrh * 1.0e9  # converting from GPa to Pascal units (kg/ms^2)
        K = self.K_vrh * 1.0e9
        return np.sqrt(((3 * K + 4 * G) / (3.0 * self.structure.density)))

    @property
    def velocity_average(self):
        """


        Returns
        -------
        float
            Average Sound velocity(m/s)

        """
        vt = self.velocity_transverse
        vl = self.velocity_logitudinal

        return 1.0 / (
            np.cbrt((1.0 / 3.0) * (2.0 / (vt * vt * vt) + 1.0 / (vl * vl * vl)))
        )

    @property
    def debye_temperature(self):
        """


        Returns
        -------
        theta : float
            Debye temperature calculated using  Orson Anderson's proposal [Ref- J. Phys. Chem. Solids (1963) 24, 909-917].
            WARNING: Debye model for the atomic displacement is based on a monoatomic crystal, here we consider an average mass if your crystal has several species

        """
        if self.structure is None:
            raise Exception("No structure was provided")
        q = self.structure.natoms
        density = self.structure.density
        total_mass = np.sum(self.structure.masses)
        theta = (
            (h_Planck / kB)
            * self.velocity_average
            * np.cbrt((3 * q * N_avogadro * density) / (4 * (np.pi) * total_mass))
        )
        return theta

    @property
    def melting_temperature(self):
        """


        Returns
        -------
        float
                Melting temperature estimated using empirical relation from Ref: Johnston I, Keeler G, Rollins R and Spicklemire S 1996
                Solid State Physics Simulations, The Consortium for Upper-Level Physics Software (New York: Wiley)

        """
        return 607 + 9.3 * self.K_vrh

    def print_properties(self):

        print("\n \n                         Voigt     Reuss    Average")
        print("-------------------------------------------------------")
        print(
            "Bulk modulus   (GPa)  %9.3f %9.3f %9.3f "
            % (self.K_v, self.K_r, self.K_vrh)
        )
        print(
            "Shear modulus  (GPa)  %9.3f %9.3f %9.3f "
            % (self.G_v, self.G_r, self.G_vrh)
        )
        print(
            "Young modulus  (GPa)  %9.3f %9.3f %9.3f "
            % (self.E_v, self.E_r, self.E_vrh)
        )
        print(
            "Poisson ratio         %9.3f %9.3f %9.3f "
            % (self.Nu_v, self.Nu_r, self.Nu_vrh)
        )
        print(
            "P-wave modulus  (GPa) %9.3f %9.3f %9.3f "
            % (self.M_v, self.M_r, self.M_vrh)
        )
        print(
            "Bulk/Shear ratio      %9.3f %9.3f %9.3f (%s) "
            % (self.KG_ratio_v, self.KG_ratio_r, self.KG_ratio_vrh, self.ductility)
        )
        print("-------------------------------------------------------")

        print(" \n \n \t \t Elastic Anisotropy \n ")
        print("Zener anisotropy (true for cubic crystals only) Az = %10.5f" % self.A_z)
        print(
            "Universal anisotropy index (Ranganathan and Ostoja-Starzewski method; PRL 101, 055504 (2008)) Au = %10.5f"
            % self.A_u
        )
        print(
            "Log-Euclidean anisotropy parameter by Christopher M. Kube, AIP Advances 6, 095209 (2016) AL = %10.5f "
            % self.A_l
        )
        print(
            "\n \t \t  Elastic wave velocities calculated using Navier's equation  (in m/s units) \n"
        )
        print("----------------------------------------------- ")
        print("Longitudinal wave velocity (vl) : %10.5f " % self.velocity_logitudinal)
        print("Transverse wave velocity (vt) : %10.5f " % self.velocity_transverse)
        print("Average wave velocity (vm) : %10.5f " % self.velocity_average)
        print("Debye temperature  (in K) : %10.5f " % self.debye_temperature)
        print(
            "WARNING: Debye model for the atomic displacement is based on a monoatomic crystal, here we consider an average mass in case your crystal has several species."
        )
        # print(
        #     "Atomic displacement at 150, 300 and 450 K  (in A^2) : %10.5f %10.5f %10.5f"
        #     % (
        #         u2FromDebye(mass, natoms, theta, 150.0),
        #         u2FromDebye(mass, natoms, theta, 300.0),
        #         u2FromDebye(mass, natoms, theta, 450.0),
        #     )
        # )
        print(
            "\nMelting temperature calculated from empirical relation: Tm = 607 + 9.3*Kvrh \pm 555 (in K)"
        )
        print("Tm (in K)=  %10.5f (plus-minus 555 K) " % self.melting_temperature)
        print("------------------------------------------------------------------ ")

        print("Hardness analysis from 6 different semi-empirical relationships.")
        H1a, H1b, H2, H3, H4, H5 = self.hardness
        print("------------------------------------------------------------------ ")
        print("Hardness (H1a) =", "{:.2f}".format(H1a), "GPa", "  Ref.[1]")
        print("Hardness (H1b) =", "{:.2f}".format(H1b), "GPa", "  Ref.[1]")
        print("Hardness (H2)  =", "{:.2f}".format(H2), "GPa", "  Ref.[2]")
        print("Hardness (H3)  =", "{:.2f}".format(H3), "GPa", "  Ref.[3]")
        print("Hardness (H4)  =", "{:.2f}".format(H4), "GPa", "  Ref.[4]")
        print("Hardness (H5)  =", "{:.2f}".format(H5), "GPa", "  Ref.[5]")
        print("References:")
        print(
            "[1] Correlation between hardness and elastic moduli of the covalent crystals. Jiang, et al. (2011)."
        )
        print(
            "[2] Computational alchemy: the search for new superhard materials. Teter (1998)."
        )
        print(
            "[3] Mechanical and electronic properties of B12-based ternary crystals of orthorhombic phase. Jiang et al. (2010)."
        )
        print(
            "[4] Theoretical investigation on the transition-metal borides with Ta3B4-type structure: A class of hard and refractory materials. Miao et al. (2011)."
        )
        print(
            "[5] Modeling hardness of polycrystalline materials and bulk metallic glasses. Chen et al. (2011)."
        )

    @property
    def elastic_stability(self):
        return eigenvals.positive_evals(self.elastic_tensor)

    @property
    def hardness(self):
        """
        Returns
        -------
        float
            The hardness calculated by 6 different methods:
            [H1a and H1b] Correlation between hardness and elastic moduli of the covalent crystals. Jiang, et al. (2011).
            [H2] Computational alchemy: the search for new superhard materials. Teter (1998).
            [H3] Mechanical and electronic properties of B12-based ternary crystals of orthorhombic phase. Jiang et al. (2010).
            [H4] Theoretical investigation on the transition-metal borides with Ta3B4-type structure: A class of hard and refractory materials. Miao et al. (2011).
            [H5] Modeling hardness of polycrystalline materials and bulk metallic glasses. Chen et al. (2011).
        """
        B = self.K_vrh
        G = self.G_vrh
        Y = self.E_vrh
        v = self.Nu_vrh
        k = G / B
        H1a = (1 / 6.78) * G
        H1b = (1 / 16.48) * Y
        H2 = (0.1769 * G) - 2.899
        H3 = (1 / 15.76) * Y
        H4 = ((1 - 2 * v) * B) / (6 * (1 + v))
        H5 = 2 * ((k * k * G) ** 0.585) - 3
        return H1a, H1b, H2, H3, H4, H5



#!/usr/bin/env python

import numpy as np


def stability_test(matrix, crystal_type):

    """This methods tests for the stability of a structure."""

    c = np.copy(matrix)

    if crystal_type == "cubic":
        print("Cubic crystal system \n")
        print(
            "Born stability criteria for the stability of cubic system are: [Ref- Mouhat and Coudert, PRB 90, 224104 (2014)]  \n"
        )
        print("(i) C11 - C12 > 0;    (ii) C11 + 2C12 > 0;   (iii) C44 > 0 \n ")

        # check (i)   keep in mind list starts with 0, so c11 is stored as c00
        if c[0][0] - c[0][1] > 0.0:
            print("Condition (i) satified.")
        else:
            print("Condition (i) NOT satisfied.")

        if c[0][0] + 2 * c[0][1] > 0.0:
            print("Condition (ii) satified.")
        else:
            print("Condition (ii) NOT satisfied.")

        if c[3][3] > 0.0:
            print("Condition (iii) satified.")
        else:
            print("Condition (iii) NOT satisfied.")

    if crystal_type == "hexagonal":
        print("Hexagonal crystal system \n")
        print(
            "Born stability criteria for the stability of hexagonal system are: [Ref- Mouhat and Coudert, PRB 90, 224104 (2014)]  \n"
        )
        print(
            "(i) C11 - C12 > 0;    (ii) 2*C13^2 < C33(C11 + C12);   (iii) C44 > 0 \n "
        )

        # check (i)   keep in mind list starts with 0, so c11 is stored as c00
        if c[0][0] - c[0][1] > 0.0:
            print("Condition (i) satified.")
        else:
            print("Condition (i) NOT satisfied.")

        if 2 * (c[0][2] * c[0][2]) < c[2][2] * (c[0][0] + c[0][1]):
            print("Condition (ii) satified.")
        else:
            print("Condition (ii) NOT satisfied.")

        if c[3][3] > 0.0:
            print("Condition (iii) satified.")
        else:
            print("Condition (iii) NOT satisfied.")

    if crystal_type == "tetragonal":
        print("Tetragonal crystal system \n")
        print(
            "Born stability criteria for the stability of Tetragonal system are: [Ref- Mouhat and Coudert, PRB 90, 224104 (2014)]  \n"
        )
        print(
            "(i) C11 - C12 > 0;    (ii) 2*C13^2 < C33(C11 + C12);   (iii) C44 > 0;   (iv) C66 > 0;    (v) 2C16^2 < C66*(C11-C12) \n "
        )

        # check (i)   keep in mind list starts with 0, so c11 is stored as c00
        if c[0][0] - c[0][1] > 0.0:
            print("Condition (i) is satified.")
        else:
            print("Condition (i) is NOT satisfied.")

        if 2 * (c[0][2] * c[0][2]) < c[2][2] * (c[0][0] + c[0][1]):
            print("Condition (ii) is satified.")
        else:
            print("Condition (ii) is NOT satisfied.")

        if c[3][3] > 0.0:
            print("Condition (iii) is satified.")
        else:
            print("Condition (iii) is NOT satisfied.")

        if c[5][5] > 0.0:
            print("Condition (iv) is satified.")
        else:
            print("Condition (iv) is NOT satisfied.")

        if 2 * c[0][5] * c[0][5] < c[5][5] * (c[0][0] - c[0][1]):
            print("Condition (v) is satified.")
        else:
            print("Condition (v) is NOT satisfied.")

    if crystal_type == "rhombohedral-1":
        print(
            "Rhombohedral (class-1): i.e. structures with point group: 3m, -3m and 32 \n"
        )
        print(
            "Born stability criteria for the stability of Rhombohedral-1 class system are: [Ref- Mouhat and Coudert, PRB 90, 224104 (2014)]  \n"
        )
        print(
            "(i) C11 - C12 > 0;    (ii) C13^2 < (1/2)*C33(C11 + C12);   (iii) C14^2 < (1/2)*C44*(C11-C12) = C44*C66;   (iv)  C44 > 0; \n "
        )

        if c[0][0] - c[0][1] > 0.0:
            print("Condition (i) is satified.")
        else:
            print("Condition (i) is NOT satisfied.")

        if (c[0][2] * c[0][2]) < (0.5) * c[2][2] * (c[0][0] + c[0][1]):
            print("Condition (ii) is satified.")
        else:
            print("Condition (ii) is NOT satisfied.")

        if c[0][3] * c[0][3] < 0.5 * c[3][3] * (c[0][0] - c[0][1]):
            print("Condition (iii) is satified.")
        else:
            print("Condition (iii) is NOT satisfied.")

        if c[3][3] > 0.0:
            print("Condition (iv) is satified.")
        else:
            print("Condition (iv) is NOT satisfied.")

    if crystal_type == "rhombohedral-2":
        print("Rhombohedral (class-2): i.e structures with point group: 3, -3 \n")
        print(
            "Born stability criteria for the stability of Rhombohedral-1 class system are: [Ref- Mouhat and Coudert, PRB 90, 224104 (2014)]  \n"
        )
        print(
            "(i) C11 - C12 > 0;    (ii) C13^2 < (1/2)*C33(C11 + C12);   (iii) C14^2 + C15^2 < (1/2)*C44*(C11-C12) = C44*C66;   (iv)  C44 > 0;  Note: C15 is added.. \n "
        )

        if c[0][0] - c[0][1] > 0.0:
            print("Condition (i) is satified.")
        else:
            print("Condition (i) is NOT satisfied.")

        if c[0][2] * c[0][2] < (0.5) * c[2][2] * (c[0][0] + c[0][1]):
            print("Condition (ii) is satified.")
        else:
            print("Condition (ii) is NOT satisfied.")

        if c[0][3] * c[0][3] + c[0][4] * c[0][4] < 0.5 * c[3][3] * (c[0][0] - c[0][1]):
            print("Condition (iii) is satified.")
        else:
            print("Condition (iii) is NOT satisfied.")

        if c[3][3] > 0.0:
            print("Condition (iv) is satified.")
        else:
            print("Condition (iv) is NOT satisfied.")

    if crystal_type == "orthorhombic":
        print("Orthorhombic crystal system.... \n")
        print(
            "Born stability criteria for the stability of Orthorhombic systems are: [Ref- Mouhat and Coudert, PRB 90, 224104 (2014)]  \n"
        )
        print(
            "(i) C11 > 0;   (ii) C11*C22 > C12^2;   (iii) C11*C22*C33 + 2C12*C13*C23 - C11*C23^2 - C22*C13^2 - C33*C12^2 > 0;   (iv)  C44 > 0;   (v)  C55 > 0 ;   (vi)  C66 > 0 \n"
        )

        # check (i)   keep in mind list starts with 0, so c11 is stored as c00
        if c[0][0] > 0.0:
            print("Condition (i) is satified.")
        else:
            print("Condition (i) is NOT satisfied.")

        if c[0][0] * c[1][1] > c[0][1] * (c[0][1]):
            print("Condition (ii) is satified.")
        else:
            print("Condition (ii) is NOT satisfied.")

        if (
            c[0][0] * c[1][1] * c[2][2]
            + 2 * c[0][1] * c[0][2] * c[1][2]
            - c[0][0] * c[1][2] * c[1][2]
            - c[1][1] * c[0][2] * c[0][2]
            - c[2][2] * c[0][1] * c[0][1]
            > 0
        ):
            print("Condition (iii) is satified.")
        else:
            print("Condition (iii) is NOT satisfied.")

        if c[3][3] > 0.0:
            print("Condition (iv) is satified.")
        else:
            print("Condition (iv) is NOT satisfied.")

        if c[4][4] > 0.0:
            print("Condition (iv) is satified.")
        else:
            print("Condition (iv) is NOT satisfied.")

        if c[5][5] > 0.0:
            print("Condition (iv) is satified.")
        else:
            print("Condition (iv) is NOT satisfied.")

    if crystal_type == "monoclinic":
        print("Monoclinic crystal system.... \n")
        print(
            "Born stability criteria for the stability of monoclinic systems are: [Ref- Mouhat and Coudert, PRB 90, 224104 (2014), and Wu et al. PRB 76, 054115 (2007)]  \n"
        )
        print(
            "(i) C11 > 0;  (ii)  C22 > 0; (iii)  C33 > 0; (iv)  C44 > 0;   (v)  C55 > 0 ;   (vi)  C66 > 0  "
        )
        print(
            " (vii) [C11 + C22 + C33 + 2*(C12 + C13 + C23)] > 0;    (viii)  C33*C55 - C35^2 > 0;   (ix)  C44*C66 - C46^2 > 0;   (x) C22 + C33 - 2*C23  > 0 "
        )
        print(
            " (xi) C22*(C33*C55 - C35^2) + 2*C23*C25*C35 - (C23^2)*C55 - (C25^2)*C33   > 0  "
        )
        print(
            " (xii)  2*[C15*C25*(C33*C12 - C13*C23) + C15*C35*(C22*C13 - C12*C23) + C25*C35*(C11*C23 - C12*C13)] - [C15*C15*(C22*C33 - C23^2) + C25*C25*(C11*C33 - C13^2) + C35*C35*(C11*C22 - C12^2)] + C55*g > 0  "
        )
        print(
            "          where, g = [C11*C22*C33 - C11*C23*C23 - C22*C13*C13 - C33*C12*C12 + 2*C12*C13*C23 ] "
        )

        if c[0][0] > 0.0:
            print("Condition (i) is satified.")
        else:
            print("Condition (i) is NOT satified.")

        if c[1][1] > 0.0:
            print("Condition (ii) is satified.")
        else:
            print("Condition (ii) is NOT satified.")

        if c[2][2] > 0.0:
            print("Condition (iii) is satified.")
        else:
            print("Condition (iii) is NOT satified.")

        if c[3][3] > 0.0:
            print("Condition (iv) is satified.")
        else:
            print("Condition (iv) is NOT satified.")

        if c[4][4] > 0.0:
            print("Condition (v) is satified.")
        else:
            print("Condition (v) is NOT satified.")

        if c[5][5] > 0.0:
            print("Condition (vi) is satified.")
        else:
            print("Condition (vi) is NOT satified.")

        # for i in range(0, 6):
        #    if(c[i][i]  > 0.0):
        #        print("Condition (%2d) is satified." % (i+1))
        #    else:
        #        print("Condition (%2d) is NOT satified." % (i+1))

        if c[0][0] + c[1][1] + c[2][2] + 2 * (c[0][1] + c[0][2] + c[1][2]) > 0:
            print("Condition (vii) is satified.")
        else:
            print("Condition (vii) is NOT satisfied.")

        if c[2][2] * c[4][4] - c[2][4] * c[2][4] > 0:
            print("Condition (viii) is satified.")
        else:
            print("Condition (viii) is NOT satisfied.")

        if c[3][3] * c[5][5] - c[3][5] * c[3][5] > 0.0:
            print("Condition (ix) is satified.")
        else:
            print("Condition (ix) is NOT satisfied.")

        if c[1][1] + c[2][2] - 2 * c[1][2] > 0.0:
            print("Condition (x) is satified.")
        else:
            print("Condition (x) is NOT satisfied.")

        if (
            c[1][1] * (c[2][2] * c[4][4] - c[2][4] * c[2][4])
            + 2 * c[1][2] * c[1][4] * c[2][4]
            - c[1][2] * c[1][2] * c[4][4]
            - c[1][4] * c[1][4] * c[2][2]
            > 0.0
        ):
            print("Condition (xi) is satified.")
        else:
            print("Condition (xi) is NOT satisfied.")

        g = (
            (c[0][0] * c[1][1] * c[2][2])
            - (c[0][0] * c[1][2] * c[1][2])
            - (c[1][1] * c[0][2] * c[0][2])
            - (c[2][2] * c[0][1] * c[0][1])
            + 2.0 * (c[0][1] * c[0][2] * c[1][2])
        )

        h1 = 2 * (
            c[0][4] * c[1][4] * (c[2][2] * c[0][1] - c[0][2] * c[1][2])
            + c[0][4] * c[2][4] * (c[1][1] * c[0][2] - c[0][1] * c[1][2])
            + c[1][4] * c[2][4] * (c[0][0] * c[1][2] - c[0][1] * c[0][2])
        )

        h2 = (
            c[0][4] * c[0][4] * (c[1][1] * c[2][2] - c[1][2] * c[1][2])
            + c[1][4] * c[1][4] * (c[0][0] * c[2][2] - c[0][2] * c[0][2])
            + c[2][4] * c[2][4] * (c[0][0] * c[1][1] - c[0][1] * c[0][1])
        )

        x = h1 - h2 + c[4][4] * g

        #  print 'x = %10.4f' % x
        #  print 'g = %10.4f' % g

        if x > 0.0:
            print("Condition (xii) is satified.")
        else:
            print("Condition (xii) is NOT satisfied.")

    return



#!/usr/bin/env python

from numpy import linalg as LA


def positive_evals(cnew):
    """This method checks the postivity of the eigenvalues
    of a matrix."""

    print("\nEigen Values of the matrix CNEW:")
    evals = LA.eigvals(cnew)
    print(evals)
    check = 0
    for i in range(len(evals)):
        if evals[i] > 0.0:
            pass
        else:
            check = 1
    if check == 1:
        print(
            "ATTENTION: One or more eigen values are negative indicating elastic instability."
        )
    if check == 0:
        print("All eigen values are positive indicating elastic stability.")

    return not(bool(check))



from .comms import printer
from .parsers import VaspOutcar
from .parsers import AbinitOutput
from .core import ELATE


def calculate_elastic_anisotropy(
    infile="OUTCAR", code="vasp", plot=None, elastic_calc=None, ddbfile=None
):

    """
    This method calculates the elastic properties
    of a material from a DFT calculation. hi
    """

    # welcome message
    printer.print_mechelastic()

    elastic_tensor = None

    rowsList = []
    # calling parser
    if code == "vasp":

        output = VaspOutcar(infile=infile)
        elastic_tensor = output.elastic_tensor
        elastic_tensor = output.elastic_tensor
        row = elastic_tensor.shape[0]
        col = elastic_tensor.shape[1]
        rowsList = []
        for i in range(row):
            columnsList = []
            for j in range(col):
                columnsList.append(elastic_tensor[i, j])
            rowsList.append(columnsList)

    elif code == "abinit":
        output = AbinitOutput(infile=infile, ddbfile=ddbfile)
        elastic_tensor = output.elastic_tensor
        elastic_tensor = output.elastic_tensor
        row = elastic_tensor.shape[0]
        col = elastic_tensor.shape[1]
        rowsList = []
        for i in range(row):
            columnsList = []
            for j in range(col):
                columnsList.append(elastic_tensor[i, j])
            rowsList.append(columnsList)

    elastic_tensor = ELATE.ELATE(rowsList)

    if plot == "2D":
        elastic_tensor.plot_2D(elastic_calc=elastic_calc)
    elif plot == "3D":
        elastic_tensor.plot_3D(elastic_calc=elastic_calc)

    elastic_tensor.print_properties()

    print("\nThanks! See you later. ")
    return output



# -*- coding: utf-8 -*-

import json
import math
import os
import platform
import random
import re
import sys
import time

from collections import OrderedDict
from io import StringIO
import requests

import numpy as np
from scipy import optimize


__author__ = "Romain Gaillac and François-Xavier Coudert"
__version__ = "2019.01.09"
__license__ = "MIT"




def make3DPlot(func, legend = '', npoints = 200):
    
      str1 = legend.split("\'")[0]
      str2 = legend.split("\'")[1]
    
      u = np.linspace(0, np.pi, npoints)
      v = np.linspace(0, 2*np.pi, 2*npoints)
      r = np.zeros(len(u)*len(v))
    
      dataX = [[0.0 for i in range(len(v))] for j in range(len(u))]
      dataY = [[0.0 for i in range(len(v))] for j in range(len(u))]
      dataZ = [[0.0 for i in range(len(v))] for j in range(len(u))]
      dataR = [["0.0" for i in range(len(v))] for j in range(len(u))]
    
      count = 0
      for cu in range(len(u)):
        for cv in range(len(v)):
          r_tmp = func(u[cu],v[cv])
          z = r_tmp * np.cos(u[cu])
          x = r_tmp * np.sin(u[cu]) * np.cos(v[cv])
          y = r_tmp * np.sin(u[cu]) * np.sin(v[cv])
          dataX[cu][cv] = x
          dataY[cu][cv] = y
          dataZ[cu][cv] = z
          dataR[cu][cv] = "'E = "+str(float(int(10*r_tmp))/10.0)+" GPa, "+"\u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"
          count = count+1
      return ((dataX,dataY,dataZ,dataR))


def make3DPlotPosNeg(func, legend = '', npoints = 200):

  u = np.linspace(0, np.pi, npoints)
  v = np.linspace(0, 2*np.pi, 2*npoints)

  dataX1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR1 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  count = 0
  for cu in range(len(u)):
    for cv in range(len(v)):
      r_tmp = max(0,func(u[cu],v[cv]))
      z = r_tmp * np.cos(u[cu])
      x = r_tmp * np.sin(u[cu]) * np.cos(v[cv])
      y = r_tmp * np.sin(u[cu]) * np.sin(v[cv])
      dataX1[cu][cv] = x
      dataY1[cu][cv] = y
      dataZ1[cu][cv] = z
      dataR1[cu][cv] = "'"+"\u03B2 = "+str(float(int(10*r_tmp))/10.0)+" TPa'"+"+'-1'.sup()+"+"', \u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"
      count = count+1

  dataX2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR2 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  count = 0
  for cu in range(len(u)):
    for cv in range(len(v)):
      r_tmp = max(0,-func(u[cu],v[cv]))
      z = r_tmp * np.cos(u[cu])
      x = r_tmp * np.sin(u[cu]) * np.cos(v[cv])
      y = r_tmp * np.sin(u[cu]) * np.sin(v[cv])
      dataX2[cu][cv] = x
      dataY2[cu][cv] = y
      dataZ2[cu][cv] = z
      dataR2[cu][cv] = "'"+"\u03B2 = -"+str(float(int(10*r_tmp))/10.0)+" TPa'"+"+'-1'.sup()+"+"', \u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"
      count = count+1

  return ((dataX1,dataY1,dataZ1,dataR1),(dataX2,dataY2,dataZ2,dataR2))

def make3DPlot2(func, legend = '', npoints = 50):

  u = np.linspace(0, np.pi, npoints)
  v = np.linspace(0, np.pi, npoints)
  w = [v[i]+np.pi for i in range(1,len(v))]
  v = np.append(v, w)

  dataX1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR1 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  dataX2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR2 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  count = 0
  r = [0.0,0.0,np.pi/2.0,np.pi/2.0]
  for cu in range(len(u)):
    for cv in range(len(v)):

      r = func(u[cu],v[cv],r[2],r[3])
      z = np.cos(u[cu])
      x = np.sin(u[cu]) * np.cos(v[cv])
      y = np.sin(u[cu]) * np.sin(v[cv])

      r1_tmp = r[0]
      z1 = r1_tmp * z
      x1 = r1_tmp * x
      y1 = r1_tmp * y
      dataX1[cu][cv] = x1
      dataY1[cu][cv] = y1
      dataZ1[cu][cv] = z1
      dataR1[cu][cv] = "'"+"G'"+"+'min'.sub()+"+"' = "+str(float(int(10*r1_tmp))/10.0)+"GPa, "+"\u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"

      r2_tmp = r[1]
      z2 = r2_tmp * z
      x2 = r2_tmp * x
      y2 = r2_tmp * y
      dataX2[cu][cv] = x2
      dataY2[cu][cv] = y2
      dataZ2[cu][cv] = z2
      dataR2[cu][cv] = "'"+"G'"+"+'max'.sub()+"+"' = "+str(float(int(10*r1_tmp))/10.0)+"GPa, "+"\u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"
      count = count+1

  i = random.randint(0, 100000)
  return ((dataX1,dataY1,dataZ1,dataR1),(dataX2,dataY2,dataZ2,dataR2))


def make3DPlot3(func, legend = '', width = 600, height = 600, npoints = 50):

  str1 = legend.split("\'")[0]
  str2 = legend.split("\'")[1]

  u = np.linspace(0, np.pi, npoints)
  v = np.linspace(0, np.pi, npoints)
  w = [v[i]+np.pi for i in range(1,len(v))]
  v = np.append(v, w)

  dataX1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR1 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  dataX2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR2 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  dataX3 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY3 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ3 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR3 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  count = 0
  r = [0.0, 0.0, 0.0, np.pi/2.0, np.pi/2.0]
  ruv = [[r for i in range(len(u))] for j in range(len(v))]
  for cu in range(len(u)):
    for cv in range(len(v)):
       ruv[cv][cu] = func(u[cu],v[cv],r[3],r[4])

  for cu in range(len(u)):
    for cv in range(len(v)):

      z = np.cos(u[cu])
      x = np.sin(u[cu]) * np.cos(v[cv])
      y = np.sin(u[cu]) * np.sin(v[cv])

      r = ruv[cv][cu]
      r1_tmp = r[0]
      dataX1[cu][cv] = r1_tmp * x
      dataY1[cu][cv] = r1_tmp * y
      dataZ1[cu][cv] = r1_tmp * z
      dataR1[cu][cv] = "'"+"\u03BD'"+"+'min'.sub()+"+"' = "+str(float(int(100*r1_tmp))/100.0)+", "+"\u03B8 = "+str(float(int(100*u[cu]*180/np.pi))/100.0)+"\u00B0, "+"\u03c6 = "+str(float(int(100*v[cv]*180/np.pi))/100.0)+"\u00B0'"

      r2_tmp = r[1]
      dataX2[cu][cv] = r2_tmp * x
      dataY2[cu][cv] = r2_tmp * y
      dataZ2[cu][cv] = r2_tmp * z
      dataR2[cu][cv] = float(int(100*r2_tmp))/100.0 

      r3_tmp = r[2]
      dataX3[cu][cv] = r3_tmp * x
      dataY3[cu][cv] = r3_tmp * y
      dataZ3[cu][cv] = r3_tmp * z
      dataR3[cu][cv] = "'"+"\u03BD'"+"+'max'.sub()+"+"' = "+str(float(int(100*r3_tmp))/100.0)+", "+"\u03B8 = "+str(float(int(100*u[cu]*180/np.pi))/100.0)+"\u00B0, "+"\u03c6 = "+str(float(int(100*v[cv]*180/np.pi))/100.0)+"\u00B0'"
      count = count+1

  return ((dataX1,dataY1,dataZ1,dataR1),(dataX2,dataY2,dataZ2,dataR2),(dataX3,dataY3,dataZ3,dataR3))




# Polar plot functions
################################################################################################


def makePolarPlot(func, legend = '', p="xy", npoints = 90):
  u = np.linspace(0, 2*np.pi, npoints)
  r = list(map(func, u))
  if (p=="xy"):
    x = r * np.cos(u)
    y = r * np.sin(u)
  else:
    y = r * np.cos(u)
    x = r * np.sin(u)

  return ((x,y))

def makePolarPlotPosNeg(func, legend = '', p="xy", npoints = 90):
  
  u = np.linspace(0, 2*np.pi, npoints)
  r = list(map(lambda x: max(0, func(x)), u))
  if (p=="xy"):
    x1 = r * np.cos(u)
    y1 = r * np.sin(u)
  else:
    y1 = r * np.cos(u)
    x1 = r * np.sin(u)
  r = list(map(lambda x: max(0, -func(x)), u))
  if (p=="xy"):
    x2 = r * np.cos(u)
    y2 = r * np.sin(u)
  else:
    y2 = r * np.cos(u)
    x2 = r * np.sin(u)

  return ((x1,y1),(x2,y2))

def makePolarPlot2(func, legend = '', p="xy", npoints = 61):


  u = np.linspace(0, 2*np.pi, npoints)
  r = list(map(func, u))

  if (p=="xy"):
    x1 = np.array([ ir[0] * np.cos(iu) for ir, iu in zip(r,u) ])
    y1 = np.array([ ir[0] * np.sin(iu) for ir, iu in zip(r,u) ])
    x2 = np.array([ ir[1] * np.cos(iu) for ir, iu in zip(r,u) ])
    y2 = np.array([ ir[1] * np.sin(iu) for ir, iu in zip(r,u) ])
  else:
    y1 = np.array([ ir[0] * np.cos(iu) for ir, iu in zip(r,u) ])
    x1 = np.array([ ir[0] * np.sin(iu) for ir, iu in zip(r,u) ])
    y2 = np.array([ ir[1] * np.cos(iu) for ir, iu in zip(r,u) ])
    x2 = np.array([ ir[1] * np.sin(iu) for ir, iu in zip(r,u) ])

  return ((x1,y1),(x2,y2))

def makePolarPlot3(func, legend = '', p="xy", npoints = 61):


  u = np.linspace(0, 2*np.pi, npoints)
  r = list(map(func, u))

  if (p=="xy"):
    x1 = np.array([ ir[0] * np.cos(iu) for ir, iu in zip(r,u) ])
    y1 = np.array([ ir[0] * np.sin(iu) for ir, iu in zip(r,u) ])
    x2 = np.array([ ir[1] * np.cos(iu) for ir, iu in zip(r,u) ])
    y2 = np.array([ ir[1] * np.sin(iu) for ir, iu in zip(r,u) ])
    x3 = np.array([ ir[2] * np.cos(iu) for ir, iu in zip(r,u) ])
    y3 = np.array([ ir[2] * np.sin(iu) for ir, iu in zip(r,u) ])
  else:
    y1 = np.array([ ir[0] * np.cos(iu) for ir, iu in zip(r,u) ])
    x1 = np.array([ ir[0] * np.sin(iu) for ir, iu in zip(r,u) ])
    y2 = np.array([ ir[1] * np.cos(iu) for ir, iu in zip(r,u) ])
    x2 = np.array([ ir[1] * np.sin(iu) for ir, iu in zip(r,u) ])
    y3 = np.array([ ir[2] * np.cos(iu) for ir, iu in zip(r,u) ])
    x3 = np.array([ ir[2] * np.sin(iu) for ir, iu in zip(r,u) ])

  return ((x1,y1),(x2,y2),(x3,y3))


########################################################################################################

def dirVec(theta, phi):
  return [ math.sin(theta)*math.cos(phi), math.sin(theta)*math.sin(phi), math.cos(theta) ]

def dirVec1(theta, phi, chi):
  return [ math.sin(theta)*math.cos(phi), math.sin(theta)*math.sin(phi), math.cos(theta) ]

def dirVec2(theta, phi, chi):
  return [ math.cos(theta)*math.cos(phi)*math.cos(chi) - math.sin(phi)*math.sin(chi),
          math.cos(theta)*math.sin(phi)*math.cos(chi) + math.cos(phi)*math.sin(chi),
          - math.sin(theta)*math.cos(chi) ]


# Functions to minimize/maximize
def minimize(func, dim):
  if dim == 2:
    r = ((0, np.pi), (0, np.pi))
    n = 25
  elif dim == 3:
    r = ((0, np.pi), (0, np.pi), (0, np.pi))
    n = 10

  # TODO -- try basin hopping or annealing
  return optimize.brute(func, r, Ns = n, full_output = True, finish = optimize.fmin)[0:2]

def maximize(func, dim):
  res = minimize(lambda x: -func(x), dim)
  return (res[0], -res[1])


class ELATE:
    def __init__(self, s):
        
        self.elas = Elastic(s)
        self.elasList = s
        
        
        
        minE = minimize(self.elas.Young, 2)
        maxE = maximize(self.elas.Young, 2)
        minLC = minimize(self.elas.LC, 2)
        maxLC = maximize(self.elas.LC, 2)
        minG = minimize(self.elas.shear, 3)
        maxG = maximize(self.elas.shear, 3)
        minNu = minimize(self.elas.Poisson, 3)
        maxNu = maximize(self.elas.Poisson, 3)
        
        self.voigtE = self.elas.averages()[0][0]
        self.reussE = self.elas.averages()[1][0]
        self.hillE  = self.elas.averages()[2][0]
        self.max_E = maxE[1]
        self.min_E = minE[1]
        self.min_axis_E = tuple(dirVec(*minE[0]))
        self.max_axis_E = tuple(dirVec(*maxE[0]))
        self.anis_E = maxE[1]/minE[1]
        
        self.voigtLC = self.elas.averages()[0][1]
        self.reussLC = self.elas.averages()[1][1]
        self.hillLC  = self.elas.averages()[2][1]
        self.max_LC = maxLC[1]
        self.min_LC = minLC[1]
        self.min_axis_LC = tuple(dirVec(*minLC[0]))
        self.max_axis_LC = tuple(dirVec(*maxLC[0]))
        if minLC[1] > 0:
            self.anis_LC = maxLC[1]/minLC[1]
        else:
            self.anis_LC = "&infin;"
            
        self.voigtShear = self.elas.averages()[0][2]
        self.reussShear = self.elas.averages()[1][2]
        self.hillShear  = self.elas.averages()[2][2]
        self.max_Shear = maxG[1]
        self.min_Shear = minG[1]
        self.min_axis_Shear = tuple(dirVec1(*minG[0]))
        self.max_axis_Shear = tuple(dirVec1(*maxG[0]))
        self.mix_2nd_axis_Shear = tuple(dirVec2(*minG[0]))
        self.max_2nd_axis_Shear = tuple(dirVec2(*maxG[0]))
        self.anis_Shear = maxG[1]/minG[1]
        
        self.voigtPoisson = self.elas.averages()[0][3]
        self.reussPoisson = self.elas.averages()[1][3]
        self.hillPoisson  = self.elas.averages()[2][3]
        self.max_Poisson = maxNu[1]
        self.min_Poisson = minNu[1]
        self.min_axis_Poisson = tuple(dirVec1(*minNu[0]))
        self.max_axis_Poisson = tuple(dirVec1(*maxNu[0]))
        self.min_2nd_axis_Poisson = tuple(dirVec2(*minNu[0]))
        self.max_2nd_axis_Poisson = tuple(dirVec2(*maxNu[0]))
        if minNu[1]*maxNu[1] > 0:
            self.anis_Poisson = maxNu[1]/minNu[1]
        else:
            self.anix_Poisson = "&infin;"
            
            
        
    def YOUNG2D(self):
        data1 = makePolarPlot(lambda x: self.elas.Young([np.pi / 2, x]), "Young's modulus in (xy) plane", "xy")
        data2 = makePolarPlot(lambda x: self.elas.Young([x, 0]), "Young's modulus in (xz) plane", "xz")
        data3 = makePolarPlot(lambda x: self.elas.Young([x, np.pi / 2]), "Young's modulus in (yz) plane", "yz")
        return (data1,data2,data3)
    
    def YOUNG3D(self):
    
          if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)
        
        
          data = make3DPlot(lambda x, y: self.elas.Young_2(x, y), "Young's modulus")
        
          return data
    
    def LC2D(self):
        data1 = makePolarPlotPosNeg(lambda x: self.elas.LC([np.pi / 2, x]), "linear compressibility in (xy) plane", "xy")
        data2 = makePolarPlotPosNeg(lambda x: self.elas.LC([x, 0]), "linear compressibility in (xz) plane", "xz")
        data3 = makePolarPlotPosNeg(lambda x: self.elas.LC([x, np.pi / 2]),  "linear compressibility in (yz) plane", "yz")
        return (data1,data2,data3)
    
    def LC3D(self):
    
          if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)
        
          data = make3DPlotPosNeg(lambda x, y: self.elas.LC_2(x, y), "Linear compressiblity")
        
          return data
    
    def SHEAR2D(self):
        data1 = makePolarPlot2(lambda x: self.elas.shear2D([np.pi / 2, x]), "Shear modulus in (xy) plane", "xy")
        data2 = makePolarPlot2(lambda x: self.elas.shear2D([x, 0]), "Shear modulus in (xz) plane", "xz")
        data3 = makePolarPlot2(lambda x: self.elas.shear2D([x, np.pi / 2]), "Shear modulus in (yz) plane", "yz")
        return (data1,data2,data3)
    
    def SHEAR3D(self):
    
          if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)
        
          data = make3DPlot2(lambda x, y, g1, g2: self.elas.shear3D(x, y, g1, g2), "Shear modulus")
        
        
          return data
    
    def POISSON2D(self):
        data1 = makePolarPlot3(lambda x: self.elas.Poisson2D([np.pi / 2, x]), "Poisson's ratio in (xy) plane", "xy")
        data2 = makePolarPlot3(lambda x: self.elas.Poisson2D([x, 0]), "Poisson's ratio in (xz) plane", "xz")
        data3 = makePolarPlot3(lambda x: self.elas.Poisson2D([x, np.pi / 2]), "Poisson's ratio in (yz) plane", "yz")
        return (data1,data2,data3)
    
    def POISSON3D(self):
    
          if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)
        
          data = make3DPlot3(lambda x, y, g1, g2: self.elas.poisson3D(x, y, g1, g2), "Poisson's ratio")
        
          return data
    
    
    ##############################################################################
    # Plotting functions
    #############################################################################
    def plot_3D(self,elastic_calc = ''):
        import pyvista as pv 
        from pyvistaqt import BackgroundPlotter
        
        plotter = BackgroundPlotter()
        
        x = None
        y = None
        z = None
        r = None
        
        if elastic_calc == 'POISSON':
            func = self.POISSON3D()
            colors = ['red','green','blue']
            for ix,icolor in zip(range(len(func)),colors):
                x = np.array(func[ix][0])
                y = np.array(func[ix][1])
                z = np.array(func[ix][2])
                r = np.array(func[ix][3])
                if np.all((func[ix][0] == 0)):
                    continue
                else:
                    grid = pv.StructuredGrid(x, y, z)
                    if (ix == 2):
                        plotter.add_mesh(grid,opacity = 0.25, color = icolor )
                    else:
                        plotter.add_mesh(grid, opacity = 0.50, color = icolor )
                    
        elif elastic_calc == 'SHEAR':
            func = self.SHEAR3D()
            colors = ['green','blue']
            for ix,icolor in zip(range(len(func)),colors):
                x = np.array(func[ix][0])
                y = np.array(func[ix][1])
                z = np.array(func[ix][2])
                r = np.array(func[ix][3])
                if np.all((func[ix][0] == 0)):
                    continue
                else:
                    grid = pv.StructuredGrid(x, y, z)
                    if (ix == 2):
                        plotter.add_mesh(grid,opacity = 0.25, color = icolor )
                    else:
                        plotter.add_mesh(grid, opacity = 0.50, color = icolor )
                        
        elif elastic_calc == 'LC':
            func = self.LC3D()
            colors = ['green','red']
            for ix,icolor in zip(range(len(func)),colors):
                x = np.array(func[ix][0])
                y = np.array(func[ix][1])
                z = np.array(func[ix][2])
                r = np.array(func[ix][3])
                if np.all((func[ix][0] == 0)):
                    continue
                else:
                    grid = pv.StructuredGrid(x, y, z)
                    if (ix == 2):
                        plotter.add_mesh(grid,opacity = 0.25, color = icolor )
                    else:
                        plotter.add_mesh(grid, opacity = 0.50, color = icolor )
    
        elif elastic_calc == 'YOUNG':
            func = self.YOUNG3D()
            colors = ['green']
            for ix,icolor in zip(range(len(func)),colors):
                x = np.array(func[ix][0])
                y = np.array(func[ix][1])
                z = np.array(func[ix][2])
                r = np.array(func[ix][3])
                if np.all((func[ix][0] == 0)):
                    continue
                else:
                    grid = pv.StructuredGrid(x, y, z)
                    if (ix == 2):
                        plotter.add_mesh(grid,opacity = 0.25, color = icolor )
                    else:
                        plotter.add_mesh(grid, opacity = 0.50, color = icolor )
                        
        plotter.add_axes()
        plotter.show_grid()
        
    def plot_2D(self,elastic_calc =''):
        import matplotlib.pyplot as plt
        
        fig = plt.figure()
        subTitles = ['XY','XZ','YZ']
        
        if elastic_calc == 'POISSON':
            func = self.POISSON2D()
            colors = ['red','green','blue']
            labels = ['Poisson - Neg', 'Poisson - Pos','Poisson - Max']
            fig.suptitle('Poisson\'s ratio')
            for iplane,title in zip(range(len(func)),subTitles):
                
                ax = fig.add_subplot(1,3,iplane+1)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                for iplot,color in zip(range(len(func[0])),colors):
                    plt.plot(func[iplane][iplot][0],func[iplane][iplot][1], color = color)
                    
        elif elastic_calc == 'SHEAR':
            func = self.SHEAR2D()
            colors = ['green','blue']
            labels = ['Shear - Max', 'Shear -']
            fig.suptitle('Shear modulus')
            for iplane,title in zip(range(len(func)),subTitles):
                
                ax = fig.add_subplot(1,3,iplane+1)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                for iplot,color in zip(range(len(func[0])),colors):
                    plt.plot(func[iplane][iplot][0],func[iplane][iplot][1], color = color)
        
        elif elastic_calc == 'LC':
            func = self.LC2D()
            colors = ['green','red']
            labels = ['LC-positive','LC-negative']
            fig.suptitle('Linear Compression')
            for iplane,title in zip(range(len(func)),subTitles):
            
                ax = fig.add_subplot(1,3,iplane+1)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                for iplot,color,label in zip(range(len(func[0])),colors,labels):
                    plt.plot(func[iplane][iplot][0],func[iplane][iplot][1], color = color,label = label)
    
                    
        elif elastic_calc == 'YOUNG':
            func = self.YOUNG2D()
            color = 'green'
            label = 'YOUNG'
            fig.suptitle('Young\'s Modulus')
            for iplane, title in zip(range(len(func)),subTitles):
                ax = fig.add_subplot(1,3,iplane+1)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                plt.plot(func[iplane][0],func[iplane][1], color = color, label = label)
       
    
    def print_properties(self):
#        print("\n \n Input: compliance matrix(coefficients in GPa) of\n")
#        for x in self.elasList:
#            print(str(x))
            
        print("\n")
        print("\n")
        print("Average properties")
        
        print("\n \n                         Voigt     Reuss    Hill")
        print("-------------------------------------------------------")
        print(
            "Bulk modulus   (GPa)  %9.3f %9.3f %9.3f "
            % (self.elas.averages()[0][0], self.elas.averages()[1][0], self.elas.averages()[2][0])
        )
        print(
            "Shear modulus  (GPa)  %9.3f %9.3f %9.3f "
            % (self.elas.averages()[0][1], self.elas.averages()[1][1], self.elas.averages()[2][1])
        )
        print(
            "Young modulus  (GPa)  %9.3f %9.3f %9.3f "
            % (self.elas.averages()[0][2], self.elas.averages()[1][2], self.elas.averages()[2][2])
        )
        print(
            "Poisson ratio         %9.3f %9.3f %9.3f "
            % (self.elas.averages()[0][3], self.elas.averages()[1][3], self.elas.averages()[2][3]))
        
        print("-------------------------------------------------------")
        #printer.printMatix(elastic_rowsList)
        print("\n")
        print("Eigenvalues of compliance matrix")
        print("\n \n   lamda_1  lamda_2  lamda_3  lamda_4  lamda_5  lamda_6")    
        print("---------------------------------------------------------------")
        eigenval = sorted(np.linalg.eig(self.elas.CVoigt)[0])
        print("%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f" % tuple(eigenval))
        print("--------------------------------------------------------------")
        if eigenval[0] <= 0:
            print('<div class="error">Stiffness matrix is not definite positive, crystal is mechanically unstable<br/>')
        print("\n")
        print("Variations of the elastic moduli")
#        print("\n \n   Young's Modulus                Linear Compressibility                   Shear Modulus              Poisson's Ratio")
#        print("    Emin         Emax             betaMin            betaMax               Gmin        Gmax           Nu_min      Nu_max")
#        print("----------------------------------------------------------------------------------------------")
        minE = minimize(self.elas.Young, 2)
        maxE = maximize(self.elas.Young, 2)
        minLC = minimize(self.elas.LC, 2)
        maxLC = maximize(self.elas.LC, 2)
        minG = minimize(self.elas.shear, 3)
        maxG = maximize(self.elas.shear, 3)
        minNu = minimize(self.elas.Poisson, 3)
        maxNu = maximize(self.elas.Poisson, 3)
        #print(type((minE,maxE,minLC,maxLC,minG,maxG,minNu,maxNu)))
        #print("%9.3f GPa  %9.3f GPa  %9.3f TPa^-1   %9.3f TPa^-1    %9.3f GPa  %9.3f GPa  %9.3f  %9.3f" % (minE[1],maxE[1],minLC[1],maxLC[1],minG[1],maxG[1],minNu[1],maxNu[1]))
        
        anisE = '%8.4f' % (maxE[1]/minE[1])
        anisLC = ('%8.4f' % (maxLC[1]/minLC[1])) if minLC[1] > 0 else "&infin;"
        anisG = '%8.4f' % (maxG[1]/minG[1])
        anisNu = ('%8.4f' % (maxNu[1]/minNu[1])) if minNu[1]*maxNu[1] > 0 else "&infin;"
        
        #print("       %s                     %s                                %s                         %s" % (anisE,anisLC,anisG,anisNu))
        #print(dirVec1(*minG[0]))
        
        print("\n \n                                  Min       Max   ||   Anisotropy")
        print("----------------------------------------------------------------------------------------")
        print(
            "Young's Modulus     (GPa)      %9.3f %9.3f || %s "
            % (minE[1], maxE[1], anisE)
        )
        print(
            "         Min Axis:    %s      \n         Max Axis:    %s "
            % (str(tuple(dirVec(*minE[0]))), str(tuple(dirVec(*maxE[0]))))
        )
        print("----------------------------------------------------------------------------------------")   
        print(
            "Linear Compression   (TPa^-1)  %9.3f %9.3f || %s "
            % (minE[1], maxE[1], anisLC)
        )
        print(
            "         Min Axis:    %s      \n         Max Axis:    %s "
            % (str(tuple(dirVec(*minLC[0]))), str(tuple(dirVec(*maxLC[0]))))
        )
        print("----------------------------------------------------------------------------------------")   
        print(
            "Shear Modulus   (GPa)          %9.3f %9.3f || %s "
            % (minE[1], maxE[1], anisG)
        )
        print(
            "         Min Axis:    %s      \n         Max Axis:    %s "
            % (str(tuple(dirVec1(*minG[0]))), str(tuple(dirVec1(*maxG[0]))))
        )
        print(
            "         Second Min Axis:    %s      \n         Second Max Axis:    %s "
            % (str(tuple(dirVec2(*minG[0]))), str(tuple(dirVec2(*maxG[0]))))
        )
        print("----------------------------------------------------------------------------------------")   
        print(
            "Poisson Ratio                 %9.3f %9.3f || %s "
            % (minE[1], maxE[1], anisNu)
        )
        print(
            "         Min Axis:    %s      \n         Max Axis:    %s "
            % (str(tuple(dirVec1(*minNu[0]))), str(tuple(dirVec1(*maxNu[0]))))
        )
        print(
            "         Second Min Axis:    %s      \n         Second Max Axis:    %s "
            % (str(tuple(dirVec2(*minNu[0]))), str(tuple(dirVec2(*maxNu[0]))))
        )
#        print(
#            "Shear modulus  (GPa)  %9.3f %9.3f %9.3f "
#            % (self.elas.averages()[0][1], self.elas.averages()[1][1], self.elas.averages()[2][1])
#        )
#        print(
#            "Young modulus  (GPa)  %9.3f %9.3f %9.3f "
#            % (self.elas.averages()[0][2], self.elas.averages()[1][2], self.elas.averages()[2][2])
#        )
#        print(
#            "Poisson ratio         %9.3f %9.3f %9.3f "
#            % (self.elas.averages()[0][3], self.elas.averages()[2][3], self.elas.averages()[2][3]))
#        
#        print("-------------------------------------------------------")
#        print("%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f" % tuple(dirVec(*minE[0])) + tuple(dirVec(*maxE[0])) + tuple(dirVec(*minLC[0])) + tuple(dirVec(*maxLC[0])) + tuple(dirVec1(*minG[0])) +  tuple(dirVec1(*maxG[0])) + tuple(dirVec1(*minNu[0])) + tuple(dirVec1(*maxNu[0])))
#        print("%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f" % tuple(dirVec2(*minG[0])) + tuple(dirVec2(*maxG[0])) + tuple(dirVec2(*minNu[0])) + tuple(dirVec2(*maxNu[0])))



class Elastic:
  """An elastic tensor, along with methods to access it"""

  def __init__(self, s):
    """Initialize the elastic tensor from a string"""
    
    if not s:
      raise ValueError("no matrix was provided")

    # Argument can be a 6-line string, a list of list, or a string representation of the list of list
    try:
      if type(json.loads(s)) == list: s = json.loads(s)
    except:
      pass

    if type(s) == str:
      # Remove braces and pipes
      s = s.replace("|", " ").replace("(", " ").replace(")", " ")

      # Remove empty lines
      lines = [line for line in s.split('\n') if line.strip()]
      if len(lines) != 6:
        raise ValueError("should have six rows")

      # Convert to float
      try:
        mat = [list(map(float, line.split())) for line in lines]
      except:
        raise ValueError("not all entries are numbers")
    elif type(s) == list:
      # If we already have a list, simply use it
      mat = s
    else:
      raise ValueError("invalid argument as matrix")

    # Make it into a square matrix
    mat = np.array(mat)
    if mat.shape != (6,6):
      # Is it upper triangular?
      if list(map(len, mat)) == [6,5,4,3,2,1]:
        mat = [ [0]*i + mat[i] for i in range(6) ]
        mat = np.array(mat)

      # Is it lower triangular?
      if list(map(len, mat)) == [1,2,3,4,5,6]:
        mat = [ mat[i] + [0]*(5-i) for i in range(6) ]
        mat = np.array(mat)

    if mat.shape != (6,6):
      raise ValueError("should be a square matrix")

    # Check that is is symmetric, or make it symmetric
    if np.linalg.norm(np.tril(mat, -1)) == 0:
      mat = mat + np.triu(mat, 1).transpose()
    if np.linalg.norm(np.triu(mat, 1)) == 0:
      mat = mat + np.tril(mat, -1).transpose()
    if np.linalg.norm(mat - mat.transpose()) > 1e-3:
      raise ValueError("should be symmetric, or triangular")
    elif np.linalg.norm(mat - mat.transpose()) > 0:
      mat = 0.5 * (mat + mat.transpose())

    # Store it
    self.CVoigt = mat

    # Put it in a more useful representation
    try:
      self.SVoigt = np.linalg.inv(self.CVoigt)
    except:
      raise ValueError("matrix is singular")

    VoigtMat = [[0, 5, 4], [5, 1, 3], [4, 3, 2]]
    def SVoigtCoeff(p,q): return 1. / ((1+p//3)*(1+q//3))

    self.Smat = [[[[ SVoigtCoeff(VoigtMat[i][j], VoigtMat[k][l]) * self.SVoigt[VoigtMat[i][j]][VoigtMat[k][l]]
                     for i in range(3) ] for j in range(3) ] for k in range(3) ] for l in range(3) ]
    return

  def isOrthorhombic(self):
    def iszero(x): return (abs(x) < 1.e-3)
    return (iszero(self.CVoigt[0][3]) and iszero(self.CVoigt[0][4]) and iszero(self.CVoigt[0][5])
            and iszero(self.CVoigt[1][3]) and iszero(self.CVoigt[1][4]) and iszero(self.CVoigt[1][5])
            and iszero(self.CVoigt[2][3]) and iszero(self.CVoigt[2][4]) and iszero(self.CVoigt[2][5])
            and iszero(self.CVoigt[3][4]) and iszero(self.CVoigt[3][5]) and iszero(self.CVoigt[4][5]))

  def isCubic(self):
      def iszero(x): return (abs(x) < 1.e-3)
      return (iszero(self.CVoigt[0][3]) and iszero(self.CVoigt[0][4]) and iszero(self.CVoigt[0][5])
                and iszero(self.CVoigt[1][3]) and iszero(self.CVoigt[1][4]) and iszero(self.CVoigt[1][5])
                and iszero(self.CVoigt[2][3]) and iszero(self.CVoigt[2][4]) and iszero(self.CVoigt[2][5])
                and iszero(self.CVoigt[3][4]) and iszero(self.CVoigt[3][5]) and iszero(self.CVoigt[4][5])
                and iszero(self.CVoigt[0][0] - self.CVoigt[1][1]) and iszero(self.CVoigt[0][0] - self.CVoigt[2][2])
                and iszero(self.CVoigt[0][0] - self.CVoigt[1][1]) and iszero(self.CVoigt[0][0] - self.CVoigt[2][2])
                and iszero(self.CVoigt[3][3] - self.CVoigt[4][4]) and iszero(self.CVoigt[3][3] - self.CVoigt[5][5])
                and iszero(self.CVoigt[0][1] - self.CVoigt[0][2]) and iszero(self.CVoigt[0][1] - self.CVoigt[1][2]))

  def Young(self, x):
        a = dirVec(x[0], x[1])
        r = sum([ a[i]*a[j]*a[k]*a[l] * self.Smat[i][j][k][l]
                  for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        return 1/r

  def Young_2(self,x,y):
        a = dirVec(x, y)
        r = sum([ a[i]*a[j]*a[k]*a[l] * self.Smat[i][j][k][l]
                  for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        return 1/r

  def LC(self, x):
        a = dirVec(x[0], x[1])
        r = sum([ a[i]*a[j] * self.Smat[i][j][k][k]
                  for i in range(3) for j in range(3) for k in range(3) ])
        return 1000 * r

  def LC_2(self, x, y):
        a = dirVec(x, y)
        r = sum([ a[i]*a[j] * self.Smat[i][j][k][k]
                  for i in range(3) for j in range(3) for k in range(3) ])
        return 1000 * r

  def shear(self, x):
        a = dirVec(x[0], x[1])
        b = dirVec2(x[0], x[1], x[2])
        r = sum([ a[i]*b[j]*a[k]*b[l] * self.Smat[i][j][k][l]
                  for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        return 1/(4*r)

  def Poisson(self, x):
        a = dirVec(x[0], x[1])
        b = dirVec2(x[0], x[1], x[2])
        r1 = sum([ a[i]*a[j]*b[k]*b[l] * self.Smat[i][j][k][l]
                  for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        r2 = sum([ a[i]*a[j]*a[k]*a[l] * self.Smat[i][j][k][l]
                  for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        return -r1/r2

  def averages(self):
        A = (self.CVoigt[0][0] + self.CVoigt[1][1] + self.CVoigt[2][2]) / 3
        B = (self.CVoigt[1][2] + self.CVoigt[0][2] + self.CVoigt[0][1]) / 3
        C = (self.CVoigt[3][3] + self.CVoigt[4][4] + self.CVoigt[5][5]) / 3
        a = (self.SVoigt[0][0] + self.SVoigt[1][1] + self.SVoigt[2][2]) / 3
        b = (self.SVoigt[1][2] + self.SVoigt[0][2] + self.SVoigt[0][1]) / 3
        c = (self.SVoigt[3][3] + self.SVoigt[4][4] + self.SVoigt[5][5]) / 3
    
        KV = (A + 2*B) / 3
        GV = (A - B + 3*C) / 5
    
        KR = 1 / (3*a + 6*b)
        GR = 5 / (4*a - 4*b + 3*c)
    
        KH = (KV + KR) / 2
        GH = (GV + GR) / 2
    
        return [ [KV, 1/(1/(3*GV) + 1/(9*KV)), GV, (1 - 3*GV/(3*KV+GV))/2],
                 [KR, 1/(1/(3*GR) + 1/(9*KR)), GR, (1 - 3*GR/(3*KR+GR))/2],
                 [KH, 1/(1/(3*GH) + 1/(9*KH)), GH, (1 - 3*GH/(3*KH+GH))/2] ]

  def shear2D(self, x):
        ftol = 0.001
        xtol = 0.01
        def func1(z): return self.shear([x[0], x[1], z])
        r1 = optimize.minimize(func1, np.pi/2.0, args=(), method = 'Powell', options={"xtol":xtol, "ftol":ftol})#, bounds=[(0.0,np.pi)])
        def func2(z): return -self.shear([x[0], x[1], z])
        r2 = optimize.minimize(func2, np.pi/2.0, args=(), method = 'Powell', options={"xtol":xtol, "ftol":ftol})#, bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun))

  def shear3D(self, x, y, guess1 = np.pi/2.0, guess2 = np.pi/2.0):
        tol = 0.005
        def func1(z): return self.shear([x, y, z])
        r1 = optimize.minimize(func1, guess1, args=(), method = 'COBYLA', options={"tol":tol})#, bounds=[(0.0,np.pi)])
        def func2(z): return -self.shear([x, y, z])
        r2 = optimize.minimize(func2, guess2, args=(), method = 'COBYLA', options={"tol":tol})#, bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun), float(r1.x), float(r2.x))

  def Poisson2D(self, x):
        ftol = 0.001
        xtol = 0.01
        def func1(z): return self.Poisson([x[0], x[1], z])
        r1 = optimize.minimize(func1, np.pi/2.0, args=(), method = 'Powell', options={"xtol":xtol, "ftol":ftol})#, bounds=[(0.0,np.pi)])
        def func2(z): return -self.Poisson([x[0], x[1], z])
        r2 = optimize.minimize(func2, np.pi/2.0, args=(), method = 'Powell', options={"xtol":xtol, "ftol":ftol})#, bounds=[(0.0,np.pi)])
        return (min(0,float(r1.fun)), max(0,float(r1.fun)), -float(r2.fun))

  def poisson3D(self, x, y, guess1 = np.pi/2.0, guess2 = np.pi/2.0):
        tol = 0.005
        def func1(z): return self.Poisson([x, y, z])
        r1 = optimize.minimize(func1, guess1, args=(), method = 'COBYLA', options={"tol":tol})#, bounds=[(0.0,np.pi)])
        def func2(z): return -self.Poisson([x, y, z])
        r2 = optimize.minimize(func2, guess2, args=(), method = 'COBYLA', options={"tol":tol})#, bounds=[(0.0,np.pi)])
        return (min(0,float(r1.fun)), max(0,float(r1.fun)), -float(r2.fun), float(r1.x), float(r2.x))


class ElasticOrtho(Elastic):
  """An elastic tensor, for the specific case of an orthorhombic system"""

  def __init__(self, arg):
        """Initialize from a matrix, or from an Elastic object"""
        if type(arg) == str:
          Elastic.__init__(self, arg)
        elif isinstance(arg, Elastic):
          self.CVoigt = arg.CVoigt
          self.SVoigt = arg.SVoigt
          self.Smat = arg.Smat
        else:
          raise TypeError("ElasticOrtho constructor argument should be string or Elastic object")

  def Young(self, x):
        ct2 = math.cos(x[0])**2
        st2 = 1 - ct2
        cf2 = math.cos(x[1])**2
        sf2 = 1 - cf2
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s44 = 4 * self.Smat[1][2][1][2]
        s55 = 4 * self.Smat[0][2][0][2]
        s66 = 4 * self.Smat[0][1][0][1]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
        return 1/(ct2**2*s33 + 2*cf2*ct2*s13*st2 + cf2*ct2*s55*st2 + 2*ct2*s23*sf2*st2 + ct2*s44*sf2*st2 + cf2**2*s11*st2**2 + 2*cf2*s12*sf2*st2**2 + cf2*s66*sf2*st2**2 + s22*sf2**2*st2**2)

  def LC(self, x):
        ct2 = math.cos(x[0])**2
        cf2 = math.cos(x[1])**2
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
        return 1000 * (ct2 * (s13 + s23 + s33) + (cf2 * (s11 + s12 + s13) + (s12 + s22 + s23) * (1 - cf2)) * (1 - ct2))

  def shear(self, x):
        ct = math.cos(x[0])
        ct2 = ct*ct
        st2 = 1 - ct2
        cf = math.cos(x[1])
        sf = math.sin(x[1])
        sf2 = sf*sf
        cx = math.cos(x[2])
        cx2 = cx*cx
        sx = math.sin(x[2])
        sx2 = 1 - cx2
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s44 = 4 * self.Smat[1][2][1][2]
        s55 = 4 * self.Smat[0][2][0][2]
        s66 = 4 * self.Smat[0][1][0][1]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
        r = (
              ct2*ct2*cx2*s44*sf2 + cx2*s44*sf2*st2*st2 + 4*cf**3*ct*cx*(-2*s11 + 2*s12 + s66)*sf*st2*sx
              + 2*cf*ct*cx*sf*(ct2*(s44 - s55) + (4*s13 - 4*s23 - s44 + s55 - 4*s12*sf2 + 4*s22*sf2 - 2*s66*sf2)*st2)*sx
              + s66*sf2*sf2*st2*sx2 + cf**4*st2*(4*ct2*cx2*s11 + s66*sx2)
              + ct2*(2*cx2*(2*s33 + sf2*(-4*s23 - s44 + 2*s22*sf2))*st2 + s55*sf2*sx2)
              + cf**2*(ct2*ct2*cx2*s55 + ct2*(-2*cx2*(4*s13 + s55 - 2*(2*s12 + s66)*sf2)*st2 + s44*sx2)
                       + st2*(cx2*s55*st2 + 2*(2*s11 - 4*s12 + 2*s22 - s66)*sf2*sx2))
            )
        return 1/r

  def Poisson(self, x):
        ct = math.cos(x[0])
        ct2 = ct*ct
        st2 = 1 - ct2
        cf = math.cos(x[1])
        sf = math.sin(x[1])
        cx = math.cos(x[2])
        sx = math.sin(x[2])
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s44 = 4 * self.Smat[1][2][1][2]
        s55 = 4 * self.Smat[0][2][0][2]
        s66 = 4 * self.Smat[0][1][0][1]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
    
        return (
      (-(ct**2*cx**2*s33*st2) - cf**2*cx**2*s13*st2*st2 - cx**2*s23*sf**2*st2*st2 + ct*cx*s44*sf*st2*(ct*cx*sf + cf*sx) -
              ct**2*s23*(ct*cx*sf + cf*sx)**2 - cf**2*s12*st2*(ct*cx*sf + cf*sx)**2 - s22*sf**2*st2*(ct*cx*sf + cf*sx)**2 +
              cf*ct*cx*s55*st2*(cf*ct*cx - sf*sx) - cf*s66*sf*st2*(ct*cx*sf + cf*sx)*(cf*ct*cx - sf*sx) -
              ct**2*s13*(cf*ct*cx - sf*sx)**2 - cf**2*s11*st2*(cf*ct*cx - sf*sx)**2 - s12*sf**2*st2*(cf*ct*cx - sf*sx)**2)/
            (ct**4*s33 + 2*cf**2*ct**2*s13*st2 + cf**2*ct**2*s55*st2 + 2*ct**2*s23*sf**2*st2 + ct**2*s44*sf**2*st2 +
              cf**4*s11*st2*st2 + 2*cf**2*s12*sf**2*st2*st2 + cf**2*s66*sf**2*st2*st2 + s22*sf**4*st2*st2)
        )
#########################################################################################
# Data genertaation
################################################################################################
# ELATE : basic usage of the tool, only 2D plots
# YOUNG3D : visualize Young's modulus in 3D
# LC3D : visualize Linear compressiblity in 3D
# SHEAR3D : visualize Shear modulus in 3D
# POISSON3D : visualize Poisson ratio in 3D
################################################################################################




      


#!/usr/bin/env python

from .comms import printer
from .parsers import VaspOutcar
from .parsers import AbinitOutput
from .core import ElasticProperties
from .core import ElasticProperties2D


def calculate_elastic(
    infile="OUTCAR", dim="3D", crystal=None, code="vasp", ddbfile=None
):

    """
    This method calculates the elastic properties
    of a material from a DFT calculation.
    """

    # welcome message
    printer.print_mechelastic()

    elastic_tensor = None
    structure = None
    lattice_constant = None
    crystal_type = crystal

    # calling parser
    if code == "vasp":
        output = VaspOutcar(infile=infile)
        elastic_tensor = output.elastic_tensor
        structure = output.structure
        lattice_constant = output.lattice_constant

    elif code == "abinit":
        output = AbinitOutput(infile=infile, ddbfile=ddbfile)
        elastic_tensor = output.elastic_tensor
        structure = output.structure
        lattice_constant = output.lattice_constant

    # elastic constants calculation for 3D materials
    if dim == "3D":
        elastic_properties = ElasticProperties(elastic_tensor, structure, crystal_type)
        elastic_properties.print_properties()

    # elastic constants calculation for 2D materials
    elif dim == "2D":
        elastic_properties = ElasticProperties2D(elastic_tensor, lattice_constant)
        elastic_properties.print_properties()

    # other
    # else: We don't need this
    #     elastic_bulk.elastic_const_bulk(
    #         cnew, snew, crystal, cell, density, natoms, totalmass
    #     )

    print("\nThanks! See you later. ")
    return output



import numpy as np
import re
from ..utils.constants import *
from ..utils.elements import ELEMENTS, elements_inversed

from ..comms import printer
from ..tests import symmetry

from ..core import ElasticProperties
from ..core import Structure


class VaspOutcar:

    """
    This class contains methods to parse the OUTCAR file.
    """

    def __init__(self, infile="OUTCAR"):

        self.infile = infile

        self.elastic_tensor = None
        self.compaliance_tensor = None
        self.structure = None
        self.text = None

        self._parse_outcar()

        # self.elastic_properties = ElasticProperties(
        #     self.elastic_tensor, self.structure)#, self.crystal_type)

        return

    def _parse_outcar(self):
        """
        This function reads outcar and initializes the class variables

        """

        rf = open(self.infile)
        self.text = rf.read()
        rf.close()
        data = self.text
        mass = np.array(
            [float(x) for x in re.findall("POMASS\s*=\s*([0-9.]*);\s*ZVAL", data)]
        )
        volume = float(re.findall("volume of cell\s*:\s*([0-9.]*)", data)[-1])
        nions = int(re.findall("NIONS\s*=\s*([0-9]*)", data)[0])
        natoms = np.array(
            [
                int(x)
                for x in re.findall("ions per type\s*=\s*([\s0-9]*)", data)[0].split()
            ]
        )

        # Sometimes LATTYP is not present.
        # Only parse if available.
        lattice_type = re.findall("LATTYP.*", data)
        if lattice_type:
            lattyp = lattice_type[-1]
        lattice = np.array(
            [
                x.split()[:3]
                for x in re.findall(
                    "direct lattice vectors.*\n([-+0-9.\s\n]*)length of vectors", data
                )[-1].split("\n")[:3]
            ]
        ).astype(float)
        positions = np.array(
            [
                x.split()[:3]
                for x in re.findall(
                    "position of ions in fractional coordinates.*\n([-+0-9.\s\n]*)",
                    data,
                )[-1].split("\n")[:nions]
            ]
        ).astype(float)
        iontype = [
            int(x)
            for x in re.findall("ions per type\s*=\s*([0-9\s]*)", data)[0].split()
        ]
        species = re.findall("VRHFIN\s*=([a-zA-Z\s]*):", data)
        self.lattice_constant = [
            float(x)
            for x in re.findall("length of vectors.*\n([0-9.\s]*)", data)[-1].split()[
                :3
            ]
        ]

        # cell parameters
        A = float(self.lattice_constant[0])
        B = float(self.lattice_constant[1])
        C = float(self.lattice_constant[2])

        print(
            "Lattice parameters (in Angs.): a = %10.5f      b = %10.5f     c = %10.5f"
            % (A, B, C)
        )

        # external pressure in kB -> GPa
        self.pressure = float(
            re.findall(r"external\s*pressure\s=\s*([-0-9.]*)\s*kB", data)[-1]
        )
        self.pressure = self.pressure / 10

        atomic_numbers = np.zeros(nions, dtype=np.int32)
        k = 0
        for i in range(len(iontype)):
            for j in range(iontype[i]):
                atomic_numbers[k] = ELEMENTS[species[i]]
                k = k + 1

        print("Atomic numbers")
        print(atomic_numbers)
        atomic_numbers = atomic_numbers
        symbols = [elements_inversed[x] for x in atomic_numbers]
        cell = (lattice, positions, atomic_numbers)

        print("Mass of atoms (in g/mol units): ")
        print((np.array(mass)))
        print("Number of atoms: %d" % np.sum(natoms))

        total_mass = 0.0
        for i in range(len(natoms)):
            total_mass = total_mass + natoms[i] * mass[i]
        print("Total mass (in g/mol): %10.4f " % total_mass)
        print("Volume of the cell (in Ang^3 units): %10.4f " % volume)

        # converting the units
        volume *= 1.0e-30  # from Angstrom to meters
        total_mass *= 1.0e-3  # from gram to kg
        density = total_mass / (volume * N_avogadro)

        print("\nDensity (in kg/m^3 units ): %10.5f" % density)
        print("External Pressure (in GPa units ): %10.5f" % self.pressure)

        c = np.matrix(
            [
                x.split()[1:]
                for x in re.findall(
                    "TOTAL ELASTIC MODULI \(kBar\).*\n.*\n.*\n([XYZ0-9.\s-]*)\n\s*-",
                    data,
                )[0].split("\n")
            ][:6]
        ).astype(float)

        # compaliance_tensor = c.I

        print("\nPrinting Cij matrix as read from OUTCAR\n")
        np.set_printoptions(precision=4, suppress=True)
        printer.printMatrix(c)

        self.elastic_tensor = c.copy()

        # Question for Shobit ?
        # for j in range(0, 6):
        #     self.elastic_tensor[3,j] = c[4,j]
        #     self.elastic_tensor[4,j] = c[5,j]
        #     self.elastic_tensor[5,j] = c[3,j]

        # ctemp = self.elastic_tensor.copy()

        # for i in range(0, 6):
        #     self.elastic_tensor[i,3] = self.elastic_tensor[i,4]
        #     self.elastic_tensor[i,4] = self.elastic_tensor[i,5]
        #     self.elastic_tensor[i,5] = ctemp[i,3]

        # Change the units of Cij from kBar to GPa
        self.elastic_tensor /= 10.0

        self.compaliance_tensor = self.elastic_tensor.I
        # print(self.elastic_tensor)
        print(
            "\n \n printing CNEW: Modified matrix in correct order (in GPa units)... \n For example- to generate input for the ELATE code [https://github.com/fxcoudert/elate] \n"
        )

        np.set_printoptions(precision=3, suppress=True)

        printer.printMatrix(self.elastic_tensor)

        print(
            (
                "\n Checking if the modified matrix CNEW is symmetric: i.e. Cij = Cji:  %10s"
                % symmetry.check_symmetric(self.elastic_tensor)
            )
        )
        self.structure = Structure(symbols, positions, lattice)

        return

    def _readOUTCAR(self):

        # s is the complaiance matrix and c is the elastic constant matrix
        s = np.zeros((6, 6))
        c = np.zeros((6, 6))

        # parsing OUTCAR

        f = open(self.infile, "r")
        lines = f.readlines()
        f.close()

        nions = -1
        mass = []
        foundcell = 0
        icell = -1
        iatom = -1
        lattice = []
        positions = []
        atomic_numbers = []
        symbols = []
        for i in lines:
            word = i.split()
            if "TOTAL ELASTIC MODULI (kBar)" in i:
                ii = lines.index(i)
            if "POMASS" in i and "ZVAL" in i:
                m = i.split()[2]
                mass.append(float(m[:-1]))
            if "volume of cell" in i:
                vol = float(i.split()[4])
            if "NIONS =" in i:
                nions = int(i.split()[11])
            if "ions per type =" in i:
                self.natoms = list(map(int, i.split()[4:]))
            if "LATTYP" in i:
                lattyp = i
            if icell > -1 and not foundcell:
                lattice.append(list(map(float, i.split()))[:3])
                icell = icell + 1
                if icell == 3:
                    foundcell = 1
            if not foundcell and "direct lattice vectors" in i:
                icell = icell + 1
            if nions > 0 and iatom > -1:
                positions.append(list(map(float, word)))
                iatom = iatom + 1
                if iatom == nions:
                    iatom = -1
            if nions > 0 and "position of ions in fractional coordinates" in i:
                iatom = iatom + 1
            if "ions per type = " in i:
                iontype = list(map(int, word[4:]))
            if "VRHFIN" in i:
                # symbols.append(i.split()[1][1:-1])
                # Modified by Uthpala to detect symbols in OUTCAR when there is a space between
                # the element symbol and ':'. E.g.- La : instead of La:
                symbols.append(re.findall(r"=([a-zA-Z\s]*):", i)[0].split()[0])
            if "length of vectors" in i:
                n = lines.index(i)
                l = lines[n + 1]
                self.lattconst = l.split()

        # cell parameters
        A = float(self.lattconst[0])
        B = float(self.lattconst[1])
        C = float(self.lattconst[2])

        print(
            "Lattice parameters (in Angs.): a = %10.5f      b = %10.5f     c = %10.5f"
            % (A, B, C)
        )

        atomic_numbers = np.zeros(nions, dtype=np.int32)
        k = 0
        for i in range(len(iontype)):
            for j in range(iontype[i]):
                atomic_numbers[k] = ELEMENTS[symbols[i]]
                k = k + 1

        print("Atomic numbers")
        print(atomic_numbers)
        lattice = np.array(lattice)
        positions = np.array(positions)
        self.cell = (lattice, positions, atomic_numbers)

        mass = np.array(mass)
        print("Mass of atoms (in g/mol units): ")
        print((np.array(mass)))
        print("Number of atoms: %d" % np.sum(self.natoms))

        self.totalmass = 0.0
        for i in range(len(self.natoms)):
            self.totalmass = self.totalmass + self.natoms[i] * mass[i]
        print("Total mass (in g/mol): %10.4f " % self.totalmass)
        print("Volume of the cell (in Ang^3 units): %10.4f " % vol)

        # converting the units
        vol = vol * 1.0e-30  # from Angstrom to meters
        self.totalmass = self.totalmass * 1.0e-3  # from gram to kg
        self.density = self.totalmass / (vol * N_avogadro)

        print("\nDensity (in kg/m^3 units ): %10.5f" % self.density)

        for i in range(0, 6):
            l = lines[ii + 3 + i]
            word = l.split()
            s[i][:] = word[1:7]
            c[i] = list(map(float, s[i][:]))

        # print c
        mc = np.matrix(c)
        mci = mc.I

        for i in range(0, 6):
            for j in range(0, 6):
                s[i][j] = mci[i, j]

        print("Printing Cij matrix as read from OUTCAR\n")
        printer.printMatrix(c)

        # Redefining the Cij matrix into the correct Voigt notation since VASP's OUTCAR has a different order
        # In VASP: Columns and rows are listed as: 1, 2, 3, 6, 4, 5
        # In this format OUTCAR's C44 values would be actually C66, C55 would be C44, and C66 would be C55.
        # OUTCAR follows the below order:
        # [C11 C12 C13 C16 C14 C15]
        # [C21 C22 C23 C26 C24 C25]
        # [C31 C32 C33 C36 C34 C35]
        # [C61 C62 C63 C66 C64 C65]
        # [C41 C42 C43 C46 C44 C45]
        # [C51 C52 C53 C56 C54 C55]

        self.cnew = np.copy(c)

        # for i in range(0,3):
        #   for j in range(0,6):
        #     cnew[i][j]=c[i][j]

        for j in range(0, 6):
            self.cnew[3, j] = c[4, j]
            self.cnew[4, j] = c[5, j]
            self.cnew[5, j] = c[3, j]

        ctemp = np.zeros((6, 6))
        ctemp = np.copy(self.cnew)

        for i in range(0, 6):
            self.cnew[i, 3] = self.cnew[i, 4]
            self.cnew[i, 4] = self.cnew[i, 5]
            self.cnew[i, 5] = ctemp[i, 3]

        # Change the units of Cij from kBar to GPa
        for i in range(0, 6):
            for j in range(0, 6):
                self.cnew[i, j] = self.cnew[i, j] / 10.0

        print(
            "\n \n printing CNEW: Modified matrix in correct order (in GPa units)... \n For example- to generate input for the ELATE code [https://github.com/fxcoudert/elate] \n"
        )

        np.set_printoptions(precision=3, suppress=True)

        printer.printMatrix(self.cnew)

        print(
            (
                "\n Checking if the modified matrix CNEW is symmetric: i.e. Cij = Cji:  %10s"
                % symmetry.check_symmetric(self.cnew)
            )
        )
